//
// Software License for AMDiS
//
// Copyright (c) 2010 Dresden University of Technology 
// All rights reserved.
// Authors: Simon Vey, Thomas Witkowski et al.
//
// This file is part of AMDiS
//
// See also license.opensource.txt in the distribution.


#include "ProblemImplicit.h"
#include "MathFunctions.h"

using namespace std;

namespace AMDiS {

  void ProblemImplicitBase::readDofVec(std::istream& in, DOFVector<double>* vec, 
                                       Mesh* mesh) 
  {
    long size = mesh->getNumberOfVertices();

    TEST_EXIT(vec->getSize()>=size)("dofvector smaller than vertex data vector\n");
    std::string buf;
    getline(in, buf);
    while (!in.eof() && buf != "vertex values: solution") 
      getline(in, buf);

    TEST_EXIT(!in.eof())("no vertex data\n");

    for (long i = 0; i < size ; ++i) {
      in >> buf;
      (*vec)[i] = atof(buf.c_str());
    }
  }


  void ProblemImplicitBase::readR(std::istream& inStream, double eps, 
	                          Mesh* mesh, int implMesh , int comp) 
  {
    FUNCNAME("ProblemImplicitBase::readR()");

    DOFVector<double>* r = getSignedDistance(implMesh, comp);
    DOFVector<double>* phi1 = getPhi1(implMesh, comp);
    DOFVector<double>* phi2 = getPhi2(implMesh, comp);
    DOFVector<double>* levelSet = getLevelset(implMesh, comp);

    TEST_EXIT(r != NULL)("no signed distance vector\n");
    TEST_EXIT(phi1 != NULL)("no phasefield1 vector\n");
    TEST_EXIT(phi2 != NULL)("no phasefield2 vector\n");
    TEST_EXIT(levelSet != NULL)("no levelSet vector\n");

    bool checkSize = r->getSize() == phi1->getSize() && 
      r->getSize() == phi2->getSize();

    TEST_EXIT(checkSize)("something went wrong\n");
    
    readDofVec(inStream, r, mesh);

    for (int i = r->getSize() - 1; i >= 0; --i) {
      (*phi1)[i] = Phi1((*r)[i], eps);
      (*phi2)[i] = Phi2((*r)[i], eps);
      (*levelSet)[i] = LevelSet((*r)[i]);
    }
  }


  void ProblemImplicitBase::readPhi1(istream& inStream, double eps, 
                                     Mesh* mesh, int implMesh, int comp)
  {
    DOFVector<double>* r = getSignedDistance(implMesh, comp);
    DOFVector<double>* phi1 = getPhi1(implMesh, comp);
    DOFVector<double>* phi2 = getPhi2(implMesh, comp);
    DOFVector<double>* levelSet = getLevelset(implMesh, comp);

    TEST_EXIT(r != NULL)("no signed distance vector\n");
    TEST_EXIT(phi1 != NULL)("no phasefield1 vector\n");
    TEST_EXIT(phi2 != NULL)("no phasefield2 vector\n");
    TEST_EXIT(levelSet != NULL)("no levelSet vector\n");

    bool checkSize = r->getSize() == phi1->getSize() && 
      r->getSize() == phi2->getSize();

    TEST_EXIT(checkSize)("something went wrong\n");

    readDofVec(inStream, phi1, mesh);

    for (int i = r->getSize() - 1; i>= 0; --i) {
      (*r)[i] = Phi1ToR((*phi1)[i], eps);
      (*phi2)[i] = 1 - (*phi1)[i];
      (*levelSet)[i] = LevelSet((*r)[i]);
    }
  }


  void ProblemImplicitBase::readPhi2(istream& inStream, double eps, 
	                             Mesh* mesh, int implMesh, int comp)
  { 
    DOFVector<double>* r = getSignedDistance(implMesh, comp);
    DOFVector<double>* phi1 = getPhi1(implMesh, comp);
    DOFVector<double>* phi2 = getPhi2(implMesh, comp);
    DOFVector<double>* levelSet = getLevelset(implMesh, comp);

    TEST_EXIT(r != NULL)("no signed distance vector\n");
    TEST_EXIT(phi1 != NULL)("no phasefield1 vector\n");
    TEST_EXIT(phi2 != NULL)("no phasefield2 vector\n");
    TEST_EXIT(levelSet != NULL)("no levelSet vector\n");

    bool checkSize = r->getSize() == phi1->getSize() &&
      r->getSize() == phi2->getSize();
    TEST_EXIT(checkSize)("something went wrong\n");

    readDofVec(inStream, phi2, mesh);

    for (int i = r->getSize() - 1; i >= 0; --i) {
      (*r)[i] = Phi2ToR((*r)[i], eps);
      (*phi1)[i] = 1 - (*phi2)[i];
      (*levelSet)[i] = LevelSet((*r)[i]);
    }
  }
  

  std::string ProblemImplicitBase::getDofFilename(std::string path, int implMesh)
  {
    std::string suffix = "[" + 
      boost::lexical_cast< std::string >(implMesh) +
      "]";

    std::string dofFilename("");
    Parameters::get(path + "dof file" + suffix, dofFilename);
    if (implMesh == 0 && dofFilename.length() == 0)
      Parameters::get(path + "dof file", dofFilename);
    return dofFilename;    
  }


  double ProblemImplicitBase::getEpsilon(std::string path, int implMesh)
  {
    std::string suffix = "[" + 
      boost::lexical_cast< std::string >(implMesh) +
      "]";

    double eps(-1);
    Parameters::get(path + "eps" + suffix, eps);
    if (implMesh == 0 && eps < 0)
      Parameters::get(path + "eps", eps);
    return eps;
  }


  int ProblemImplicitBase::getType(std::string path, int implMesh)
  {
    std::string suffix = "[" + 
      boost::lexical_cast< std::string >(implMesh) +
      "]";
    int serType(-1);
    Parameters::get(path + "type" + suffix, serType);
    if (implMesh == 0 && serType < 0)
      Parameters::get(path + "type", serType);
    return serType;
  }


  DOFVector<double>* ProblemImplicit::getSignedDistance(unsigned int im , unsigned int m) 
  { 
    if (m >= r.size() || im >= r[m].size())
      return NULL;
    return (r[m])[im]; 
  }


  DOFVector<double>* ProblemImplicit::getPhi1(unsigned int im, unsigned int m)
  {
    if (m >= phi1.size() || im >= phi1[m].size())
      return NULL;

    return (phi1[m])[im];
  }


  DOFVector<double>* ProblemImplicit::getPhi2(unsigned int im, unsigned int m)
  {
    if (m >= phi2.size() || im >= phi2[m].size())
      return NULL;

    return (phi2[m])[im];
  }


  DOFVector<double>* ProblemImplicit::getLevelset(unsigned int im, unsigned int m)
  {
    if (m >= levelSet.size() || im >= levelSet[m].size())
      return NULL;

    return (levelSet[m])[im];
  }


  bool ProblemImplicit::createImplicitMesh()
  {
    //check each mesh if it's an implicit one
    r.resize(meshes.size());
    phi1.resize(meshes.size());
    phi2.resize(meshes.size());
    levelSet.resize(meshes.size());
    for (unsigned int i = 0; i < meshes.size(); ++i)
      createImplicitMesh(i);
    return true;
  }


  bool ProblemImplicit::createImplicitMesh(int p) 
  {
    FUNCNAME("ProblemImplicit::createImplicitMesh()");

    if (!isImplicitMesh(p))
      return false;
    std::string path = name + "->implicit mesh[" 
      + boost::lexical_cast< std::string >(p) + "]->";
    int nImplMeshes(0);
    Parameters::get(path + "nr meshes", nImplMeshes);

    r[p].resize(nImplMeshes);
    phi1[p].resize(nImplMeshes);
    phi2[p].resize(nImplMeshes);
    levelSet[p].resize(nImplMeshes);

    for ( int i = 0; i < nImplMeshes ; ++i ) {
      (r[p])[i] = new DOFVector< double >(getFeSpace(p), "r");
      (phi1[p])[i] = new DOFVector< double >(getFeSpace(p), "phi1");
      (phi2[p])[i] = new DOFVector< double >(getFeSpace(p), "phi2");
      (levelSet[p])[i] = new DOFVector< double >(getFeSpace(p), "levelSet");
      createImplicitMesh(path, i, p);
    }
    return true;
  }
  

  bool ProblemImplicit::createImplicitMesh(std::string path, int implMesh, 
					   int comp)
  {
    FUNCNAME("ProblemImplicit::createImplicitMesh()");

    std::string dofFilename = getDofFilename(path, implMesh);
    if (dofFilename.length() == 0)
      return false;

    double eps = getEpsilon(path, implMesh);
    if (eps < 0.0)
      return false;
    int serType = getType(path, implMesh);
    if (serType < 0)
      return false;

    TEST_EXIT(meshes[comp] != NULL)("the mesh was not created\n");
    
    std::ifstream inStream(dofFilename.c_str());
    switch (serType) {
    case 0:
      readR(inStream, eps, meshes[comp], implMesh, comp);
      break;
    case 1:
      readPhi1(inStream, eps, meshes[comp], implMesh, comp);
      break;
    case 2:
      readPhi2(inStream, eps, meshes[comp], implMesh, comp);
      break;
    default:
      WARNING("unknown implicit mesh type\n");
    }
    inStream.close();
    return true;
  }


  void ProblemImplicit::createMesh()
  {
    FUNCNAME("ProblemImplicit::createMesh()");

    ProblemStatSeq::createMesh();

    implMesh.resize(meshes.size());
    for (unsigned int i = 0 ; i < meshes.size() ; ++i ) {
      std::string path = name + "->implicit mesh[" +
	boost::lexical_cast< std::string >(i) + "]";
      std::string meshFilename("");
      Parameters::get(path + "->macro file name", meshFilename);
      std::string serFilename("");
      Parameters::get(path + "->serialization file name", serFilename);
      implMesh[i] = true;
      if (meshFilename.length() > 0)
	meshes[i]->setName(path);
      else if (serFilename.length() > 0) {
	std::ifstream inStream(serFilename.c_str());
	meshes[i]->deserialize(inStream);
	inStream.close();
      } else
	implMesh[i] = false;
    }    
  }


  void ProblemImplicit::initialize(Flag initFlag, 
				   ProblemStatSeq* adaptProblem, 
				   Flag adoptFlag)
  {
    FUNCNAME("ProblemImplicit::initialize()");

    ProblemStatSeq::initialize(initFlag);    
    createImplicitMesh();
  }
}
